#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "misc.h"

#define ARGS_LIMIT      1024
#define BUFFER_SIZE     1024

#define err_msg(fmt, arg...) \
do { \
	errors++; \
	message(1, file_name, line_number, fmt, ##arg); \
} while (0)

#define warn_msg(fmt, arg...) \
do { \
	message(0, file_name, line_number, fmt, ##arg); \
} while (0)

enum {
	INCLUDE_LT,
	INCLUDE_DQUOTE
};

struct define_clause {
	char		*name;
	char		*replace;
	char		*file;
	unsigned int	narg;
	unsigned int	line;
};

struct buffer {
	char		*data;
	unsigned int	used;
	unsigned int	size;
};

static const char	*inc_folder;	/* a folder where to find nonlocal include files */
static const char	*file_name;     /* a file name which processes right now */
static unsigned int	line_number;    /* current line number */
static unsigned int	macros_lines;   /* how many lines macros occupies */
static unsigned int	errors;         /* counter of errors, return value */

static int		out_file_desc;  /* a descriptor where proceed lines are written */
static char		out_file_buffer[BUFFER_SIZE]; /* a temporary buffer for keeping lines */
static unsigned int	buffer_used;    /* how many bytes used in a buffer */
static struct define_clause *defs;      /* an array of all define macros */
static unsigned int	def_used;
static unsigned int	def_size;

static void process_file(const char *file);
static void include_handler(char **cur);
static void define_handler(char **cur);
static char *expand_string(char **cur, const char *end);

typedef void (*dir_handler_t)();

static struct {
	const char	*name;
	dir_handler_t	handler;
} dirs[] = {
	{ "include",	include_handler },
	{ "define",	define_handler },
	{ NULL,		NULL }
};

static void flush_buffer()
{
	ssize_t ret;
	
	ret = write_full(out_file_desc, out_file_buffer, buffer_used);
	if (ret == -1) {
		xwarning("can't write to out_file_desc fd=%d: %s",
			out_file_desc, strerror(errno));
	}
	buffer_used = 0;
}

static void write_line(const char *line)
{
	size_t len;

	return_if_fail(line != NULL);

	len = strlen(line);

	if (len+buffer_used > BUFFER_SIZE) {
		flush_buffer();
		len = len > BUFFER_SIZE ? BUFFER_SIZE : len;
		memcpy(out_file_buffer, line, len);
		buffer_used = len;
	} else {
		memcpy(out_file_buffer, line, len);
		buffer_used += len;
	}

	if (buffer_used > 0) {
		if (out_file_buffer[buffer_used-1] == '\n')
			flush_buffer();
	}
}

static void write_character(char c)
{
	ssize_t ret;

	if (c == '\n') {
		flush_buffer();
		ret = write_full(out_file_desc, &c, 1);
		if (ret == -1) {
			xwarning("can't write character to file fd=%d: %s",
				out_file_desc, strerror(errno));
		}
	} else {
		if (buffer_used+1 > BUFFER_SIZE) {
			flush_buffer();
			out_file_buffer[0] = c;
			buffer_used = 1;
		} else {
			out_file_buffer[buffer_used] = c;
			++buffer_used;
		}
	}
}

static size_t skip_ident_chars(char **cur)
{
	char *pos, *pos0;

	pos0 = pos = *cur;
	if (isalpha(*pos) || *pos == '_') {
		do {
			++pos;
		} while(isalnum(*pos) || *pos == '_');
	}
	*cur = pos;
	return pos - pos0;
}

static size_t skip_blanks(char **cur)
{
	char *pos, *pos0;

	pos0 = pos = *cur;
	while (isblank(*pos))
		++pos;
	*cur = pos;
	return pos - pos0;
}

static void remove_trailing_spaces(char *ptr, size_t len)
{
	while (len && isspace(ptr[len-1]))
		ptr[--len] = 0;
}

static void ensure_buf_len(struct buffer *buf, size_t size)
{
	if (buf->used+size >= buf->size) {
		buf->size += size + 16;
		buf->data = xrealloc(buf->data, buf->size);
	}
}

static void put_char(struct buffer *buf, const char c)
{
	ensure_buf_len(buf, 1);
	buf->data[buf->used++] = c;
}

static void put_zero(struct buffer *buf)
{
	ensure_buf_len(buf, 1);
	buf->data[buf->used++] = 0;
}

static void put_str(struct buffer *buf, const char *str)
{
	size_t str_len;

	str_len = strlen(str);
	ensure_buf_len(buf, str_len);
	memcpy(buf->data+buf->used, str, str_len);
	buf->used += str_len;
}

static void put_str_n(struct buffer *buf, const char *data, size_t size)
{
	ensure_buf_len(buf, size);
	memcpy(buf->data+buf->used, data, size);
	buf->used += size;
}

static void put_uint(struct buffer *buf, unsigned int n)
{
	char *num;
	size_t num_len;

	asprintf(&num, "%d", n);
	num_len = strlen(num);
	ensure_buf_len(buf, num_len);
	memcpy(buf->data+buf->used, num, num_len);
	buf->used += num_len;
	xfree(num);
}

static char *make_string(const char *pos, size_t len)
{
	char *str;

	str = xmalloc(len+1);
	memcpy(str, pos, len);
	str[len] = 0;

	return str;
}

static void file_begin(const char *file, unsigned int line)
{
	char *str;
	asprintf(&str, "# %d \"%s\" 1", line, file);
	write_line(str);
	write_character('\n');
	xfree(str);
}

static void file_return()
{
	char *str;
	asprintf(&str, "# %d \"%s\" 2", line_number + 1,  file_name);
	write_line(str);
	xfree(str);
}

static void include_handler(char **cur)
{
	char *file;
	char *pos, *pos1;
	const char *prev_file;
	unsigned int prev_line;
	int chr, close;
	size_t len;

	pos = *cur;
	skip_blanks(&pos);
	/* TODO: search in system-wide directory and in a current one */
	switch (*pos) {
	case '<':
		chr = INCLUDE_LT;
		close = '>';
		break;
	case '\"':
		chr = INCLUDE_DQUOTE;
		close = '\"';
		break;
	default:
		err_msg("#include expects \"FILENAME\" or <FILENAME>");
		return;
	}

	++pos; /* < or " */

	pos1 = strrchr(pos, close);
	if (pos1 == NULL) {
		err_msg("missing terminating %c character", close);
		return;
	}

	/* replace > or " with 0 */
	*pos1 = 0;

	switch (chr) {
	case INCLUDE_LT:
		len = strlen(inc_folder);
		file = xmalloc(len + pos1 - pos + 1);
		xstrncpy(file, inc_folder, len + 1);
		xstrncpy(file+len, pos, pos1 - pos + 1);
		break;
	case INCLUDE_DQUOTE:
		file = xmalloc(pos1 - pos + 1);
		xstrncpy(file, pos, pos1 - pos + 1);
		break;
	}

	/* skip > or " but now it is 0 */
	++pos1;
	pos = pos1;

	skip_blanks(&pos);
	if (*pos != 0)
		warn_msg("extra tokens at end of #include directive");

	file_begin(file, line_number);
	/* store a line number and a file name */
	prev_line = line_number;
	prev_file = file_name;
	/* FALSE because this is an include file */
	process_file(file);
	/* restore previous file name and line number*/
	file_name = prev_file;
	line_number = prev_line;
	file_return();
	xfree(file);
}

static void define_handler(char **cur)
{
	struct define_clause def;
	struct buffer replace;
	char *pos, *pos0;
	char *ident;
	char **args;
	unsigned int nargs, i;
	size_t size;

	pos = *cur;
	/* skip blanks before a macros name */
	skip_blanks(&pos);
	/* get an identifier name */
	pos0 = pos;
	size = skip_ident_chars(&pos);
	ident = make_string(pos0, size);

	if (strlen(ident) == 0) {
		err_msg("macro names must be an identifier");
		goto err1;
	}
	/* handle macros parameters */
	args = NULL;
	nargs = 0;
	if (*pos == '(') {
		++pos; /* skip ( and spaces before the first argument */
		skip_blanks(&pos);
		while (*pos != ')') {
			char *arg;
			/* get the end of the string inside brackets */
			if (*pos == 0) {
				err_msg("missing ')' in macro parameter list");
				goto err2;
			}
			/* get an argument */
			pos0 = pos;
			size = skip_ident_chars(&pos);
			arg = make_string(pos0, size);
			/* don't find an identifier, check for a comma */
			if (strlen(arg) == 0) {
				if (*pos == ',')
					err_msg("parameter name missing");
				else
					err_msg("%c may not appear in macro parameter list", *pos);
				xfree(arg);
				goto err2;
			}
			/* a macros parameter name must be unique */
			for (i = 0; i < nargs; i++) {
				if (STREQ(args[i], arg)) {
					err_msg("duplicate macro parameter \"%s\"", arg);
					xfree(arg);
					goto err2;
				}
			}
			/* skip blanks after the argument */
			skip_blanks(&pos);
			/* just two symbols are valid in this position */
			if (*pos != ')' && *pos != ',') {
				err_msg("macros parameters must be comma-separated");
				xfree(arg);
				goto err2;
			}
			if (*pos == ',') {
				++pos;
				/* skip blanks before the next argument */
				skip_blanks(&pos);
				/* this check is necessary to catch this ,) */
				if (*pos == ')') {
					err_msg("parameter name missing");
					xfree(arg);
					goto err2;
				}
			}
			args = xrealloc(args, sizeof(char *) * (nargs+1));
			args[nargs++] = arg;
			if (nargs > ARGS_LIMIT) {
				err_msg("exceed an arguments number limit");
				goto err2;
			}
		}
		/* consume round bracket */
		++pos;
	}
	/* skip blanks before a replace statement */
	skip_blanks(&pos);
	/* handle a replace statement of macros */
	replace.data = NULL;
	replace.used = replace.size = 0;
	/* while not the end of the string */
	while (*pos != 0) {
		char *buf;
		/* even if we have more than one symbol between words
		 * take only one blank symbol, so get a beautiful string */
		if (isblank(*pos)) {
			++pos;
			put_char(&replace, ' ');
			skip_blanks(&pos);
			continue;
		}

		pos0 = pos;
		size = skip_ident_chars(&pos);
		buf = make_string(pos0, size);
		/* not an identifier just copy single symbol */
		if (strlen(buf) == 0) {
			put_char(&replace, *pos);
			xfree(buf);
			++pos;
			continue;
		}
		/*
		 *  buf must be compared with argument strings and if it is
		 * the same it will be replaced with the pattern $#
		 * # - number of an argument from 0 to ARGS_LIMIT
		 *
		 */
		for (i = 0; i < nargs; i++) {
			if (STREQ(buf, args[i])) {
				/* convert a number to string */
				put_char(&replace, '$');
				put_uint(&replace, i);
				break;
			}
		}
		/* don't find in arguments, write a word as it is */
		if (i == nargs)
			put_str(&replace, buf);
		xfree(buf);
	}
	/* remove a blank symbol at the end of the string and terminate it */
	remove_trailing_spaces(replace.data, replace.used);
	put_zero(&replace);
	/* need also expand the pattern string */
	def.name = xstrdup(ident);
	def.file = xstrdup(file_name);
	def.replace = expand_string(&replace.data, replace.data + replace.used);
	def.narg = nargs;
	def.line = line_number - macros_lines;
	/* store this structure */
	if (def_used == def_size) {
		def_size += 16;
		defs = xrealloc(defs, sizeof(struct define_clause) * def_size);
	}
	/* check for macros redefinition, replace previous */
	for (i = 0 ; i < def_used; i++) {
		if (STREQ(defs[i].name, def.name)) {
			warn_msg("%s redefined, previous definition was in %s:%d",
						def.name, defs[i].file, defs[i].line);
			xfree(defs[i].name);
			xfree(defs[i].file);
			xfree(defs[i].replace);
			defs[i] = def;
		}
	}
	/* don't find previous definition this is a new one */
	if (i == def_used)
		defs[def_used++] = def;
	/* free resources */
	xfree(replace.data);
err2:
	if (args != NULL) {
		for (i = 0; i < nargs; i++)
			xfree(args[i]);
		xfree(args);
	}
err1:
	xfree(ident);
}

static char *expand_string(char **cur, const char *end)
{
	struct define_clause def;
	struct buffer str;
	char **args;
	unsigned int narg, i, size;
	char *pos, *pos0;

	pos = *cur;
	str.data = NULL;
	str.size = str.used = 0;
	narg = 0;
	/* while not the end of a string */
	while (pos < end) {
		char *buf, *chr, c;
		int inside_brackets;
		/* get all blanks */
		if (isblank(*pos)) {
			pos0 = pos;
			size = skip_blanks(&pos);
			put_str_n(&str, pos0, size);
		}
		/* open and close quote */
		if (*pos == '\"') {
			pos0 = pos;
			++pos;
			chr = strchr(pos, '\"');
			if (chr != NULL) {
				++chr;
				c = *chr;
				*chr = 0;
				put_str(&str, pos0);
				*chr = c;
				pos = chr;
			} else {
				put_str(&str, pos0);
				pos += strlen(pos); /* it should be end */
			}
			continue;
		}
		/* put an identifier into buf */
		pos0 = pos;
		size = skip_ident_chars(&pos);
		buf = make_string(pos0, size);
		/* not identifier, put a symbol as it is */
		if (strlen(buf) == 0) {
			put_char(&str, *pos);
			free(buf);
			++pos;
			continue;
		}
		/* compare an identifier with macros names */
		for (i = 0; i < def_used; i++) {
			if (STREQ(defs[i].name, buf)) {
				def = defs[i];
				break;
			}
		}
		/* don't find a macros name, put a buffer as it is */
		if (i == def_used) {
			put_str(&str, buf);
			xfree(buf);
			continue;
		}
		/* a macros and an identifier have the same name
		 * there are 2 special cases which should be handled
		 */
		if ((def.narg == 0 && *pos == '(') ||
			(def.narg != 0 && *pos != '(')) {
			put_str(&str, buf);
			xfree(buf);
			continue;
		}
		/* case of a simple macros substitution */
		if (def.narg == 0) {
			put_str(&str, def.replace);
			xfree(buf);
			continue;
		}
		xfree(buf);
		/* handle arguments */
		args = NULL;
		narg = 0;
		/* count all brackets and find the last */
		inside_brackets = 1;
		pos0 = pos + 1; /* skip ( */
		do {
			++pos;
			if (*pos == '(')
				inside_brackets++;
			else if (*pos == ')')
				inside_brackets--;
			/* find end of string */
			if (pos >= end) {
				err_msg("unterminated argument list invoking macro %s", def.name);
				goto err1;
			}
		} while (inside_brackets);
		/* terminate string with 0, replace ) with 0
		 * now we have a string with all arguments separated by commas */
		*pos = 0;
		/* go through arguments */
		while (pos0 <= pos) {
			int skip_comma;
			char *arg, *pos1;
			/* skip blanks before an argument */
			skip_blanks(&pos0);
			/* pos1 points to an argument beginning */
			pos1 = pos0;
			skip_comma = 0;
			while (*pos0 != ',' || skip_comma) {
				if (pos0 == pos)
					break;
				if (*pos0 == '(')
					skip_comma++;
				else if (*pos0 == ')')
					skip_comma--;
				++pos0;
			}
			/* replace ',' with 0 */
			*pos0 = 0;
			size = pos0 - pos1 + 1;
			arg = xmalloc(size);
			xstrncpy(arg, pos1, size);
			/* remove blanks at the end of the argument */
			remove_trailing_spaces(arg, size);
			args = xrealloc(args, sizeof(char *) * (narg+1));
			args[narg++] = arg;
			*pos0 = ','; /* restore , */
			++pos0; /* skip , or at the cycle end ) which now 0 */
		}
		*pos = ')'; /* restore ) */
		++pos; /* skip ) */

		if (def.narg != narg) {
			err_msg("macro \"%s\" requires %d arguments, but %d given",
							def.name, def.narg, narg);
			goto err2;
		}
		/* a pattern replacement */
		pos0 = def.replace;
		while ((chr = strchr(pos0, '$')) != NULL ) {
			unsigned long num;
			if (chr[1] == '$') {
				pos0 += 2;
				continue;
			} else if (!isdigit(chr[1])) {
				/* something really unexpected */
				pos0 += 2;
				continue;
			}
			/* put all before $ */
			put_str_n(&str, pos0, chr-pos0);
			++chr; /* skip $ */
			num = strtoul(chr, &pos0, 10);
			if (num > ARGS_LIMIT) {
				err_msg("exceed arguments number limit");
				goto err2;
			}
			put_str(&str, args[num]);
		}
		put_str(&str, pos0);

		if (args != NULL) {
			for (i = 0; i < narg; i++)
				xfree(args[i]);
			xfree(args);
		}
	}
	put_zero(&str);

	return str.data;

err2:
	if (args != NULL) {
		for (i = 0; i < narg; i++)
			xfree(args[i]);
		xfree(args);
	}
err1:
	if (str.data != NULL)
		xfree(str.data);
	return NULL;
}

static char *remove_double_dollar(char *str)
{
	struct buffer ret;
	char *ch;
	char *pos;

	memset(&ret, 0, sizeof(struct buffer));

	pos = str;
	while ((ch = strchr(pos, '$')) != NULL) {
		if (ch[1] != '$') {
			/* Something impossible, need print a warning message */
			++ch;
			pos = ch;
			continue;
		}
		/* Skip the first '$'. */
		++ch;
		/* Copy the string including the first '$'. */
		ensure_buf_len(&ret, ch - pos);
		put_str_n(&ret, pos, ch - pos);
		/* Skip the second '$'. */
		++ch;
		pos = ch;
	}
	/* Copy the rest of the string */
	put_str(&ret, pos);
	put_zero(&ret);

	return ret.data;
}

static void process_file(const char *file)
{
	char *begin, *end, *cur, c;
	int fd, i, inside_comment;;

	size_t length, mmap_size;
	ret_t ret;

	ret = open_file_ro(file, &fd);
	/* perhaps need to handle return value to signal the right reason of error */
	if (ret != ret_ok) {
		err_msg("can't open file: %s", file);
		return;
	}

	file_name = file;
	line_number = 0;

	ret = file_length(fd, &length);
	if (ret != ret_ok)
		xerror(ret, "file length");

	mmap_size = length;
	ret = pread(fd, &c, sizeof(c), length-1);
	if (ret == -1)
		xerror(ret_err, "read file");

	if (c != '\n') {
		warn_msg("no a new line at the end of the file");
		++mmap_size;
	}

	ret = mmap_file(fd, (void **)&begin, mmap_size);
	if (ret != ret_ok)
		xerror(ret, "mmap file");

	close_file(&fd);

	begin[mmap_size - 1] = 0;

	cur = begin;
	end = begin + mmap_size;

	inside_comment = 0;
	while (cur < end) {
		struct buffer str;
		int add_cr;
		char *pos0;
		char *str_begin, *pos;

#define COMMENT_EOL	0x1
#define COMMENT_EOC	0x2

		str.data = NULL;
		str.used = str.size = 0;
		macros_lines = 0;
		add_cr = 0;
		/* proceed a file string by string, take care about
		 * multilines macros and comments */
		pos0 = cur;
		while (*cur != '\n') {
			char *pos0;
			if (*cur == 0)
				break;
			/* remove comments */
			if (!inside_comment) {
				if (*cur == '/') {
					if (cur[1] == '/') {
						inside_comment = COMMENT_EOL;
						cur += 2;
						continue;
					} else if (cur[1] == '*') {
						inside_comment = COMMENT_EOC;
						cur += 2;
						continue;
					}
				}
			} else {
				if (inside_comment & COMMENT_EOL) {
					++cur;
					if (*cur == '\n')
						inside_comment = 0;
					continue;
				} else { /* COMMENT_EOC */
					if (*cur == '*' && cur[1] == '/') {
						inside_comment = 0;
						cur += 2;
						continue;
					}
					++cur;
					continue;
				}
			}
			/* multiline macros */
			if (*cur == '\\') {
				pos0 = cur;
				++cur; /* skip '\' */
				skip_blanks(&cur);
				if (*cur == '\n') {
					++line_number;
					++macros_lines;
					++cur; /* skip '\n' */
					add_cr = 1;
					continue;
				}
				cur = pos0;
			}
			/* put double $$ instead of $ */
			if (*cur == '$')
				put_char(&str, '$');
			put_char(&str, *cur);
			++cur;
		}
		put_zero(&str);

		++line_number;
		++cur;
		/* skip blanks before # */
		str_begin = pos = str.data;
		skip_blanks(&pos);
		/* not a macros */
		if (pos[0] != '#') {
			char *tmp_str;
			char *ret_str;

			tmp_str = expand_string(&str_begin, str_begin+strlen(str_begin)+1);
			ret_str = remove_double_dollar(tmp_str);

			write_line(ret_str);

			xfree(ret_str);
			xfree(tmp_str);

			write_character('\n');
			if (add_cr)
				write_character('\n');
			xfree(str.data);
			continue;
		}
		/* macros handling */
		++pos; /* # */
		/* skip blanks before macros name */
		skip_blanks(&pos);
		/* a first macros name symbol */
		str_begin = pos;
		skip_ident_chars(&pos);

		c = *pos;
		*pos = 0;
		for (i = 0; dirs[i].name != NULL; i++) {
			if (STREQ(str_begin, dirs[i].name)) {
				if (dirs[i].handler != NULL) {
					*pos = c;
					dirs[i].handler(&pos, end);
					write_character('\n');
					if (add_cr)
						write_character('\n');
					break;
				} else {
					xerror(ret_err, "NULL macros handler");
				}
			}
		}
		xfree(str.data);

		if (dirs[i].name == NULL)
			warn_msg("unknown macros identifier %s", str_begin);
	}

	flush_buffer();
	munmap_file((void **)&begin, length);

}

unsigned int preproc_run(const char *file, int fd, const char *inc)
{
	unsigned int i;

	return_val_if_fail(file != NULL, ret_err);
	return_val_if_fail(fd >= 0, ret_err);
	return_val_if_fail(inc != NULL, ret_err);

	inc_folder = inc;
	out_file_desc = fd;
	file_begin(file, 1);
	process_file(file);

	if (defs != NULL) {
		for (i = 0; i < def_used; i++) {
			xfree(defs[i].name);
			xfree(defs[i].file);
			xfree(defs[i].replace);
		}
		xfree(defs);
	}

	return errors;
}

